// Copyright 2019 Red Hat, Inc
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.)

package common

type TranslateOptions struct {
	FilesDir                  string // allow embedding local files relative to this directory
	NoResourceAutoCompression bool   // skip automatic compression of inline/local resources
	DebugPrintTranslations    bool   // report translations to stderr
}

type TranslateBytesOptions struct {
	TranslateOptions
	Pretty bool
	Raw    bool // encode only the Ignition config, not any wrapper
}
